# Generated by Django 3.0.5 on 2020-04-14 18:12

import private_storage.fields
import private_storage.storage.files
from django.db import migrations

import openwisp_firmware_upgrader.base.models


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0002_default_permissions'),
    ]

    operations = [
        migrations.AlterField(
            model_name='firmwareimage',
            name='file',
            field=private_storage.fields.PrivateFileField(
                storage=private_storage.storage.files.PrivateFileSystemStorage(),
                upload_to=openwisp_firmware_upgrader.base.models.get_build_directory,
                verbose_name='File',
            ),
        ),
    ]
